#!/usr/bin/env bash
#
# Script to take the backup of the essential components of Key Trustee Server and Key Trustee KMS.
# Two options to setup runtime environment on Key Trustee Server host with parcels (choose one):
# 1. Copy script to KEYTRUSTEE_SERVER_PARCEL_DIR/bin directory before running
#    (e.g. /opt/cloudera/parcels/KEYTRUSTEE_SERVER/bin)
# 2. Set KT_SERVER_INSTALL_DIR environment variable before running (takes precedence over option 1).
# Two options to setup runtime environment on KeyTrustee KMS hosts with parcels (choose one):
# 1. Copy script to KEYTRUSTEE_KMS_PARCEL_DIR/bin directory before running
#    (e.g. /opt/cloudera/parcels/KEYTRUSTEE/bin -- you may need to create the bin directory)
# 2. Set KT_KMS_INSTALL_DIR environment variable before running (takes precedence over option 1).
# On Key Trustee Server or KMS hosts with packages (this will also work for parcels but it's simpler
# to follow the steps above for parcels) you must specify both the --confdir and --output options.
# Author: Suraj Acharya <sacharya@cloudera.com>

KEYTRUSTEE_SERVER_DIR="/var/lib/keytrustee"
KEYTRUSTEE_KMS_DIR="/var/lib/kms-keytrustee"
KEYTRUSTEE_SERVER_HOME="$KEYTRUSTEE_SERVER_DIR/.keytrustee" # to match meta/keytrustee_env.sh
KEYTRUSTEE_KMS_HOME="$KEYTRUSTEE_KMS_DIR/keytrustee/.keytrustee" # doesn't match anything
DB_PORT=11381
SCRIPT_NAME='ktbackup.sh'
#SCRIPT_NAME=$(basename $0) -- replaced with previous line
HOSTNAME=$(hostname | tr '[.]' '_')

TIMESTAMP=$( date '+%Y-%m-%d_%H-%M-%S' )
CLEARTEXT=0
VERBOSE=0
QUIET=0
ROLL=-1
LOGGER_AVAILABLE=${KTBACKUP_LOGGER_AVAILABLE:=0}

LOG_FILE_EXISTS=0

log() {
  [[ $QUIET -eq 0 ]] && echo "$@" 1>&2
  [[ $LOG_FILE_EXISTS -gt 0 ]] && echo "$@" >> $LOG_FILE
  if [[ $LOGGER_AVAILABLE ]]; then
    logger -p user.notice -t $SCRIPT_NAME "$@"
  fi
}

err() {
  [[ ( $QUIET -eq 0 ) || ( $LOG_FILE_EXISTS -eq 0 ) ]] && >&2 echo "$@"
  [[ $LOG_FILE_EXISTS -gt 0 ]] && echo "ERROR: $@" >> $LOG_FILE
  if [[ $LOGGER_AVAILABLE ]]; then
    logger -p user.error -t $SCRIPT_NAME "$@"
  fi
}

die() {
  ## from http://mywiki.wooledge.org/BashFAQ/101
  local st="$?" # most recent error status
  if [[ "$1" != *[^0-9]* ]]; then
    st="$1"
    shift
  fi
  err "Exiting with failure"
  exit "$st"
}

verbose() {
  [[ $VERBOSE -gt 0 ]] && echo "$@" 1>&2
  [[ $LOG_FILE_EXISTS -gt 0 ]] && echo "INFO: $@" >> $LOG_FILE
  if [[ $LOGGER_AVAILABLE ]]; then
    logger -p user.info -t $SCRIPT_NAME "$@"
  fi
}

# tar the confdir
tar_confdir() {
  if grep -q "DB_CONNECT" "$CONFDIR/keytrustee.conf" 2>> $LOG_FILE; then
    TAR_FILE="$KTS_TAR_FILE"
  else
    grep_exit=$?
    if [[ $grep_exit -lt 2 ]]; then
      TAR_FILE="$KMS_TAR_FILE"
    else
      err "Creation of configuration archive failed."
      die $grep_exit
    fi
  fi
  log "Creating tar file from $CONFDIR."
  tar -cf $TAR_FILE $CONFDIR --exclude='*.pid' 1>>$LOG_FILE 2>&1
}

# create a datadump of the database
create_db_dump() {
  if [[ "$TAR_FILE" == "$KTS_TAR_FILE" ]]; then
    verbose "Creating the database dump at $DB_FILE."
    # When user is keytrustee it defaults to database keytrustee.
    # However, when it is a different user you need to mention the database.
    if ! pg_dump -p $DB_PORT -d keytrustee > $DB_FILE 2>> $LOG_FILE; then
      pg_dump_exit=$?
      err "Database dump failed."
      err "Are you running as the keytrustee or kms user? Did you source the environment?"
      rm $DB_FILE
      rm -f $TAR_FILE
      die $pg_dump_exit
    fi
    tar --append --file=$TAR_FILE $DB_FILE 1>>$LOG_FILE 2>&1
    log "Added database dump to tarfile."
    rm $DB_FILE
    verbose "Database dump deleted."
  fi
}

# show the help
show_help() {
	echo
	echo "Options:"
	echo " -c, --confdir=CONFDIR          specifies the keytrustee configuration directory."
	echo "                                defaults to $KEYTRUSTEE_SERVER_HOME (Key Trustee Server) or"
	echo "                                defaults to $KEYTRUSTEE_KMS_HOME (Key Trustee KMS)."
  echo " --database-port=PORT           specifies the port of the Key Trustee Server database."
	echo "                                defaults to $DB_PORT"
  echo " --gpg-recipient=GPG_RECIPIENT  specifies for which gpg user payload should be encrypted."
  echo "                                See Cloudera documentation for information about"
  echo "                                configuring GPG."
  echo " --cleartext                    set this flag if you don't want the output to be encrypted."
  echo " --roll=FILES_TO_KEEP           specifies the number of files to keep. Works only with Key Trustee Server"
  echo " --output=DIR                   specify the directory to which to write the backup."
  echo "                                defaults to $KEYTRUSTEE_SERVER_DIR (Key Trustee Server) or"
  echo "                                defaults to $KEYTRUSTEE_KMS_DIR (Key Trustee KMS)."
  echo " -q, --quiet                    suppress console log messages wherever possible"
  echo "                                and return backup tar file path on success."
  echo " --verbose                      provide additional log messages during run."

	exit 0
}

# encrypt the tar with gpg
encrypt_tar() {
  if ! gpg --always-trust -e -r "$GPG_UUID" $TAR_FILE 1>>$LOG_FILE 2>&1 ; then
    gpg_exit=$?
    err "Encryption failed. See $LOG_FILE."
    rm $TAR_FILE
    log "Cleartext tarfile at $TAR_FILE deleted."
    die $gpg_exit
  else
    if [[ $CLEARTEXT -gt 0 ]]; then
      log "Cleartext tarfile left at $TAR_FILE by user request (--cleartext)."
    else
      rm $TAR_FILE
      verbose "Cleartext tarfile at $TAR_FILE deleted."
    fi
    TAR_FILE=$TAR_FILE".gpg"
    log "Encrypted tarfile at $TAR_FILE."
  fi
}

source_env_from() {
  local ENV_VAR_NAME="$1"
  local DIR="$2"
  if [[ -d "$DIR" ]]; then
    verbose "Attempting to source environment from $ENV_VAR_NAME($DIR)."
    # DIR variable now contains absolute path to script directory
    BIN_DIR="$DIR/bin"
    ENV_FILE="$DIR/meta/keytrustee_env.sh"
    ALTERNATIVES_FILE="$DIR/meta/alternatives.json"
  else
    err "$ENV_VAR_NAME set to non-existent location: $DIR"
    die 1
  fi
}

source_env_from_cd() {
  # Reference:
  # http://stackoverflow.com/questions/59895/can-a-bash-script-tell-what-directory-its-stored-in
  local SOURCE="${BASH_SOURCE[0]}"
  local DIR="$( dirname "$SOURCE" )"
  while [[ -h "$SOURCE" ]]
  do
    SOURCE="$(readlink "$SOURCE")"
    [[ $SOURCE != /* ]] && SOURCE="$DIR/$SOURCE"
    DIR="$( cd -P "$( dirname "$SOURCE"  )" && pwd )"
  done
  DIR="$( cd -P "$( dirname "$SOURCE" )" && pwd )"
  verbose "Attempting to source environment from $DIR."
  # DIR variable now contains absolute path to script directory
  BIN_DIR="$DIR/../bin"
  ENV_FILE="$DIR/../meta/keytrustee_env.sh"
  ALTERNATIVES_FILE="$DIR/../meta/alternatives.json"
}

init_env() {
  if [ -n "$KT_SERVER_INSTALL_DIR" ]; then
    if [ -n "$KT_KMS_INSTALL_DIR" ]; then
      err "Both KT_SERVER_INSTALL_DIR and KT_KMS_INSTALL_DIR specified. Cannot backup both services at once."
      err "Specify KT_SERVER_INSTALL_DIR and KT_KMS_INSTALL_DIR on separate runs of this script."
      die 1
    fi
    source_env_from "KT_SERVER_INSTALL_DIR" "$KT_SERVER_INSTALL_DIR"
    if [ -n "$KEYTRUSTEE_HOME" ]; then
      KEYTRUSTEE_HOME="$KEYTRUSTEE_SERVER_HOME"
      verbose "Manually setting KEYTRUSTEE_SERVER_HOME to $KEYTRUSTEE_SERVER_HOME. Unexpected but will continue."
    fi
  else
    if [ -n "$KT_KMS_INSTALL_DIR" ]; then
      source_env_from "KT_KMS_INSTALL_DIR" "$KT_KMS_INSTALL_DIR"
      if [ -n "$KEYTRUSTEE_HOME" ]; then
        KEYTRUSTEE_HOME="$KEYTRUSTEE_KMS_HOME"
        verbose "Manually setting KEYTRUSTEE_KMS_HOME to $KEYTRUSTEE_KMS_HOME. Expected."
      fi
    else
      source_env_from_cd
    fi
  fi

  if [ -e "$ENV_FILE" ]; then
    source ${ENV_FILE}
    verbose "Sourced environment from ${ENV_FILE}."
  fi

  if [ -n "$KEYTRUSTEE_KP_HOME" ]; then
    unset KEYTRUSTEE_HOME
  fi

  # This check must become after previous one to allow updating of KEYTRUSTEE_HOME value
  # (which keytrustee_env.sh may overwrite)
  if [ -e "$ALTERNATIVES_FILE" ] && grep -q "kms-keytrustee" "$ALTERNATIVES_FILE"; then
    KEYTRUSTEE_HOME=`grep destination "$ALTERNATIVES_FILE"|sed 's/"destination"://'|sed 's/,//'|tr -d [:blank:]|tr -d '"'`
    KEYTRUSTEE_HOME="$KEYTRUSTEE_HOME/keytrustee/.keytrustee"
    verbose "Setting KEYTRUSTEE_HOME to $KEYTRUSTEE_HOME from $ALTERNATIVES_FILE."
  fi

  if [ -n "$KEYTRUSTEE_POSTGRES_BIN" ]; then
    PATH="${PATH}:${KEYTRUSTEE_POSTGRES_BIN}"
    verbose "Added KEYTRUSTEE_POSTGRES_BIN (${KEYTRUSTEE_POSTGRES_BIN}) to PATH."
    #log "Current path is ${PATH}."
  fi

  # Note KEYTRUSTEE_HOME is not set unless it can be determined in some explicit way.
  # There is no implicit fallback to the default location because we don't want to silently
  # backup the wrong location, providing comfort without actual protection.

  if [ -z $CONFDIR ]; then
    if [ -d "$KEYTRUSTEE_HOME" ]; then
      CONFDIR="$KEYTRUSTEE_HOME"
    else
      err "Configuration directory unspecified and could not be determined. Exiting."
      die 1
    fi
  fi
  if [ -z $OUTPUTDIR ]; then
    if [ -d "$KEYTRUSTEE_HOME" ]; then
      if ! readlink -f "$KEYTRUSTEE_HOME/.." 1>>/dev/null 2>&1; then
        OUTPUTDIR="$KEYTRUSTEE_HOME/.."
      else
        OUTPUTDIR=$(readlink -f "$KEYTRUSTEE_HOME/..")
      fi
    else
      err "Output directory unspecified and could not be determined. Exiting."
      die 1
    fi
  else
    if [ ! -d "$OUTPUTDIR" ]; then
      err "Specified output directory [$OUTPUTDIR] does not exist. Exiting."
      die 1
    fi
    if [ ! -w "$OUTPUTDIR" ]; then
      err "Specified output directory [$OUTPUTDIR] is not writeable by current user. Exiting."
      die 1
    fi
  fi
}

create_log_file() {
  LOG_FILE=$OUTPUTDIR"/kt_bak_"$HOSTNAME"_"$TIMESTAMP".log"
  log "Creating log file at $LOG_FILE."
  touch $LOG_FILE
  LOG_FILE_EXISTS=1
  BOOTSTRAP_LOG_MSG="Created log file at $LOG_FILE."
  echo "$BOOTSTRAP_LOG_MSG" >> $LOG_FILE
}

# Setting up a cron job
create_cron(){
    if crontab -l | grep "ktbackup.sh" ; then
            echo "Cron job already present. Skipping"
    else
        echo "No cron job present. Will create one"
        if [ -n "$PARCELS_ROOT" ];then
            script_loc="${PARCELS_ROOT}/KEYTRUSTEE_SERVER"
        else
            script_loc=""
        fi
        cat <(crontab -l) <(echo "@hourly  $script_loc/bin/ktbackup.sh  --cleartext  --confdir=$CONFDIR --output=$OUTPUTDIR --roll=10") | crontab -
    fi
}
# read command parameters
while [ "${1:0:1}" = "-" ]; do
        case "$1" in
                -c|--confdir=*)
                        if [ "$1" = "-c" ]; then
                                CONFDIR="$2"
                                shift 2
                        else
                            CONFDIR=`echo "$1" | sed 's/--confdir=//'`
                            shift 1
                        fi
                ;;
                --database-port=*)
                        DB_PORT=`echo "$1" | sed 's/--database-port=//'`
                        shift 1
                ;;
                --gpg-recipient=*)
                        GPG_UUID=`echo "$1" | sed 's/--gpg-recipient=//'`
                        shift 1
                ;;
                --cleartext)
                        CLEARTEXT=1
                        shift
                ;;
                -q|--quiet)
                        QUIET=1
                        shift
                ;;
                --cron)
                        CRON=1
                        shift
                ;;
                --output=*)
                        OUTPUTDIR=`echo "$1" | sed 's/--output=//'`
                        shift 1
                ;;
                --verbose)
                        VERBOSE=1
                        shift 1
                ;;
                --roll=*)
                        ROLL=`echo "$1" | sed 's/--roll=//'`
                        shift 1
                ;;
                -h|--help)
                        # function exits successfully
                        show_help
                        ;;
                *)
                        echo "$1: invalid option."
                        exit 1
                        ;;
        esac
done

PROCEED=-1

if [[ $EUID -eq 0 ]]; then
   echo "Script should not run as root." 1>&2
   echo "On Key Trustee Server, run as keytrustee user: sudo -u keytrustee ./ktbackup.sh" 1>&2
   echo "On Key Trustee KMS, run as kms user: sudo -u kms ./ktbackup.sh" 1>&2
   die 1
fi

if [[ -z "$GPG_UUID" ]]; then
  # Make sure they explicitly chose cleartext output
  if [[ "$CLEARTEXT" -gt 0 ]] ; then
    PROCEED=1
    log "Creating cleartext backup at user request."
  # You can't have the cake and eat it too. Make a decision
  else
    PROCEED=-1
    err "Either provide GPG information or set the cleartext flag."
    err "Use --help option for more information."
    die 1
  fi
else
  PROCEED=2
fi

if [[ "$PROCEED" -gt 0 ]]; then
  init_env
  create_log_file
  # Set the name of the tarfile.
  # Added Hostname for uniqueness
  KMS_TAR_FILE=$OUTPUTDIR"/kms_bak_"$HOSTNAME"_"$TIMESTAMP".tar"
  KTS_TAR_FILE=$OUTPUTDIR"/kts_bak_"$HOSTNAME"_"$TIMESTAMP".tar"
  DB_FILE=$OUTPUTDIR"/kts_bak_"$HOSTNAME"_"$TIMESTAMP".sql"
  tar_confdir
  create_db_dump
  if [[ "$PROCEED" -gt 1 ]]; then
    # Use GPG to encrypt it all.
    encrypt_tar
  else
    log "Wrote backup to $TAR_FILE."
  fi
  [[ $QUIET -eq 1 ]] && echo "$TAR_FILE"
  # Once everything has been created its a good idea to remove the excess files
  if [[ $ROLL -gt 0 ]]; then
    if [[ "$TAR_FILE" == "$KTS_TAR_FILE" ]]; then
        log "Deleting old files. Keeping only the latest $ROLL file"
        KT_BAK_LOG=$OUTPUTDIR"/kt_bak*.log"
        KT_BAK_TAR=$OUTPUTDIR"/kts_bak*.tar"
        ls -tr $KT_BAK_LOG | head -n -${ROLL} | xargs rm -vf
        ls -tr $KT_BAK_TAR | head -n -${ROLL} | xargs rm -vf
    fi
  fi
  if [[ $CRON -eq 1 ]]; then
    log "Setting the cron job"
    create_cron
  fi
  # The advantage of making the check all the time is that it will help the user in case there is a late integration.
  if [[ "$TAR_FILE" == "$KTS_TAR_FILE" ]]; then
    # Look for HSM integration
    if grep -q "HSM_ADDRESS" "$CONFDIR/keytrustee.conf" ; then
        log "KeyHSM present"
        HARDWARE=$(grep  "keyhsm.hardware"  /usr/share/keytrustee-server-keyhsm/application.properties  | awk -F "=" '{print $2}' )
        log "HSM hardware is $HARDWARE"
        #Check only for Thales (ncipher). The others are cool and backup on their own.
        if [[ $HARDWARE = ncipher ]]; then
            log "Will attempt to backup $HARDWARE keys"
            NCIPHER_KEY_DATA="/opt/nfast/kmdata/local"
            if [[ ! -d $NCIPHER_KEY_DATA ]]; then
                # The directory /opt/nfast/kmdata/local does not exist. This means the keys are some place else. Tell the customer and return.
                    err "The location $NCIPHER_KEY_DATA is not present. Will not be backing up the key material."
                    exit 0
            fi
            if [ ! "$(ls -A $NCIPHER_KEY_DATA)" ]; then
                err "The folder $NCIPHER_KEY_DATA is present. But it is empty. Either $HARDWARE hasnt been setup or something went wrong"
                exit 1
            fi
            # The folder exists and there is data present within the directory. Ill consider it safe to back it up.
            log "Will backup the file(s) present in $NCIPHER_KEY_DATA for $HARDWARE"
            tar --append --file=$TAR_FILE $NCIPHER_KEY_DATA 1>>$LOG_FILE 2>&1
            log "Added the files for Thales HSM."
        fi
    fi
  fi
  exit 0
fi
